# global imports (dont't worry, red is normal)
from db.entities.user import User
from services.user_service import UserService

from flask import request, Response
import bcrypt
import jwt

from datetime import datetime, timedelta

SIGNING_KEY = "yteNrMy6142WKwp8fKfrHkS5nlFpxtHgOXJh1ZPsOrV_gTcsO9eMY7aB7HUzRbTRO9dmZhCl3FdPtuvMe3K8aBA_wc2MmHRo8IkUIGmvUJGsAxKFClN_6oNW5fEvoeVKiL1krA-qjWbR_em-WksePgPoTsySW7QbKdi4f7cwuyK2_JZ2fQj9hDKlfJ2GzMXkKiWcfyCTr30yC6BviAFeRDD_Bpvg6znsrXr53Tq66hnwDwQ6QU7aHVu-bERblKZTYuvkSxsov6yRMEVWQoiuBITsQtIOcgSWK4Dy3BjSbqoIcKw3WG-s3wx1lTen19QbEu8vJC64e0iGeGDWT6vbtg"
TOKEN_VALIDITY_IN_DAYS = 1

def generate_token(user: User) -> str:
    '''
    creates a JWT token for a user which has the following fields:
        - username
        - created_at
        - valid_until
    '''
    created_at = datetime.now()
    valid_until = created_at + timedelta(days=1)
    return jwt.encode(
            {
                'username': user.username, 
                'created_at': str(created_at), 
                'valid_until': str(valid_until),
            }, 
            SIGNING_KEY, algorithm='HS256'
        ).decode("utf-8")

def verify(token):
    '''
    verifies the validity of a JWT token.
    performs the following tests (int this order):
        - is the JWT token parsable? (it has not been damaged + the signature is valid)
        - does the payload contain all necessary fields?
        - does the user specified by the payload exist?
        - is the expiration/creation date sound?
    '''

    try:
        payload = jwt.decode(token, SIGNING_KEY, algorithms=['HS256'])
    except:
        return Response(status = 401, response=f'Invalid JWT token')

    # check if all needed fields are in the payload
    if not "username" in payload or not "created_at" in payload or not "valid_until" in payload:
        return Response(status = 401, response=f'Invalid JWT token')
    
    try:
        UserService.get_by_username(payload["username"])
    except ValueError as e:
        # return 400 if the user does not exist
        return Response(status = 400, response=str(e))

    # check if token has already expired
    token_created_at = datetime.strptime(payload["created_at"], '%Y-%m-%d %H:%M:%S.%f')
    valid_until = datetime.strptime(payload["valid_until"], '%Y-%m-%d %H:%M:%S.%f')
    now = datetime.now()

    if now <= token_created_at or now >= valid_until:
        return Response(status = 401, response=f'Token expired')
    
    # everthing is fine
    return Response(status = 200)

def authenticate():
    '''
    takes the credentials from the user and generates a JWT token out of them
    '''
    data = request.json
    username = data["username"]

    try:
        user = UserService.get_by_credentials(username, data["password"])

        return {"token": generate_token(user)}
    except ValueError as e:
        # return 400 if the user does not exist or the password is wrong
        return Response(status = 400, response=str(e))

def delete(username):
    '''
    deletes a user from the DB. should be protected later
    '''
    try:
        UserService.delete(username)

        return Response(status = 200)
    except ValueError as e:
        # return 400 if the user already exists
        return Response(status = 400, response=str(e))

def add():
    '''
    adds a new user to the DB. expects the provided password to be plaintext i.e. it should not
    be encrypted, encoded or hashed
    '''
    data = request.json

    if not "role" in data:
        data["role"] = "u"
    
    username = data["username"]

    try :
        UserService.add(username, data["password"], data["role"])
    except ValueError as e:
        # return 400 if the user already exists
        return Response(status = 400, response=str(e))

    return Response(status=200)

def all():
    '''
    return all users stored in the DB
    '''

    users = UserService._repository.all()

    return str(users)